#include "REALS.h"
#include "REALLIB.h"
#include "functclass.h"
#include <stdio.h>
#include <math.h>



REAL fact (long i)
{
	if (i==0)
		return REAL(1);
	else
		return REAL(i)*fact(i-1);
}



REAL divide_by_fact (REAL numerator, long i)
{
	if (i<=0)
		return numerator;
	else
		return divide_by_fact(numerator/i, i-1);
}


REAL funct_approx::c(long n)
{
	REAL csofar = a(n);
	int i = n-1;

	for (i=n-1; i>=0; i--)
	{
		csofar = a(i) + (b(i)/csofar);
	}

	return csofar;
}



void funct_approx::set_x(const REAL& input)
{
	x = input;
	x2 = x*x;
}



/*
REAL funct_approx::approximate(long prec)
{
	long n = 2;
	REAL y,z;

	if (prec >= 0)
		return 0;

	y = c(0);
	z = c(1);

	if (bound(z-y, prec))
		return z;
	else
	{
		while (!bound(z-y, prec))
		{
			y = z;
			z = c(n);
			n++;
		}

		printf("%ld step(s) required to converge with precision of %ld.\n\n", n, prec);

		return z;
	}
}
*/


REAL funct_approx::approximate(long prec)
{
	long n = 0;
	REAL y,z, A0, A1, B0, B1, tempA, tempB, an, bn;

	if (prec >= 0)
		return 0;

	//printf("Working with precision of %ld.\n", prec);
	A0 = REAL(1);
	A1 = a(0);
	B0 = REAL(0);
	B1 = REAL(1);

	z = A1/B1;


	do
	{
		n++;		// increment to give approximation cn
		tempA = A1;
		tempB = B1;

                an = a(n);
		bn = b(n-1);

		A1 = an*A1 + bn*A0;
		B1 = an*B1 + bn*B0;

		A0 = tempA;
		B0 = tempB;

		y = z;
		z = A1/B1;

		//printf("Approximation %ld calculated.\n", n);

	} while (!bound(z-y, prec));

	printf("%ld step(s) required to converge with precision of %ld.\n\n", n, prec);


        return z;



}



REAL stieltjes_approx::Pade(long L, long M)
{
	if (M==0)
		return 1;
	else
	{
		//printf("C(%ld,%ld)\n", L, M);
		return C_Coeff(L, M)*power(C_Base(), (M+1)*2^L);
	}

}

REAL stieltjes_approx::Taylor_a(long i)
{
	return power(C_Base(), i)*Taylor_Coeff(i)*Taylor_Base();
}


REAL stieltjes_approx::C_Coeff(long L, long M)
{
	if (M==0)
		return 1;
	else if (M==1)
		return Taylor_Coeff(L);
	else if (L==0)
		return REAL(power(REAL(-1),long(0.5*M*(M-1))));
	else
	{
		//printf("C(%ld,%ld)\n", L, M);
		return (C_Coeff(L-1, M-1)*C_Coeff(L+1, M-1)-power(C_Coeff(L, M-1),2))/C_Coeff(L, M-2);
	}

}

REAL stieltjes_approx::Stieltjes_b(long n)
{
	return -(Pade(n+2, n+1)*Pade(n, n))/(Pade(n+1, n)*Pade(n+1, n+1));
}


REAL stieltjes_approx::Stieltjes_c(long n)
{
	return -(Pade(n+2, n+2)*Pade(n+1, n))/(Pade(n+1, n+1)*Pade(n+2, n+1));
}



REAL stieltjes_approx::a(long i)
{
	if (i==0)
		return Taylor_a(0);
	else
		return 1;
}


REAL stieltjes_approx::b(long i)
{
	if (i==0)
		return Taylor_a(1);
	else if (i%2 == 0)
		return Stieltjes_b(i/2);
  	else
		return Stieltjes_c(i/2 - 1);
}


REAL jacobi_approx::a(long i)
{
	if (i==0)
		return Taylor_a(0);
	else if (i==1)
		return 1 + Stieltjes_b(0)*x;
	else
		return 1 + (Stieltjes_c(i-2) + Stieltjes_b(i-1))*x;
}


REAL jacobi_approx::b(long i)
{
	if (i==0)
		return Taylor_a(1)*x;
	else if (i == 1)
		return -Stieltjes_b(0)*Stieltjes_c(0)*x2;
  	else
		return -Stieltjes_b(i-1)*Stieltjes_c(i-1)*x2;
}



REAL sin_approximate::a(long i)
{
	if (i==0)
		return 0;
	else if (i==1)
		return 1;
	else
		return 1 - (x2)/((2*i-2)*(2*i-1));
}


REAL sin_approximate::b(long i)
{
	if (i==0)
		return x;
	else
		return (x2)/(2*i*(2*i+1));
}



REAL sin_approximate2::Taylor_Coeff(long i)
{
	return power((-1),i) * divide_by_fact(1, i*2+1);
}

REAL sin_approximate2::C_Base()
{
	return x;
}


REAL sin_approximate2::Taylor_Base()
{
	return x;
}




REAL sin_approximate3::Taylor_Coeff(long i)
{
	return power((-1),i) * divide_by_fact(1, i*2+1);
}

REAL sin_approximate3::C_Base()
{
	return x;
}

REAL sin_approximate3::Taylor_Base()
{
	return x;
}


REAL cos_approximate::a(long i)
{
	if (i==0)
		return 0;
	else if (i==1)
		return 1;
	else
		return 1 - (x2)/((2*i-3)*(2*i-2));
}


REAL cos_approximate::b(long i)
{
	if (i==0)
		return 1;
	else
		return (x2)/((2*i-1)*2*i);
}



REAL cos_approximate2::Taylor_Coeff(long i)
{
	return  power((-1),i) * divide_by_fact(1, i*2);
}

REAL cos_approximate2::C_Base()
{
	return x;
}

REAL cos_approximate2::Taylor_Base()
{
	return 1;
}



REAL cos_approximate3::Taylor_Coeff(long i)
{
	return  power((-1),i) * divide_by_fact(1, i*2);
}

REAL cos_approximate3::C_Base()
{
	return x;
}

REAL cos_approximate3::Taylor_Base()
{
	return 1;
}


REAL exp_approximate::a(long i)
{
	if (i==0)
		return 0;
	else if (i==1)
		return REAL(1);
	else
		return 1 + x/(i-1);
}


REAL exp_approximate::b(long i)
{
	if (i==0)
		return REAL(1);
	else
		return -x/i;
}




REAL exp_approximate2::Taylor_Coeff(long i)
{
	return divide_by_fact(REAL(1), i);
}

REAL exp_approximate2::C_Base()
{
	return 1;
}


REAL exp_approximate2::Taylor_Base()
{
	return 1;
}


REAL exp_approximate3::Taylor_Coeff(long i)
{
	return divide_by_fact(REAL(1), i);
}

REAL exp_approximate3::C_Base()
{
	return 1;
}


REAL exp_approximate3::Taylor_Base()
{
	return 1;
}


REAL tan_approximate::a(long i)
{
	if (i==0)
		return 0;
	else
		return 1;
}


REAL tan_approximate::b(long i)
{
	if (i==0)
		return x;
	else
		return -x2/((2*i-1)*(2*i+1));

}

REAL tanh_approximate::a(long i)
{
	if (i==0)
		return 0;
	else
		return 1;
}


REAL tanh_approximate::b(long i)
{
	if (i==0)
		return x;
	else
		return x2/((2*i-1)*(2*i+1));

}

REAL atan_approximate::a(long i)
{
	if (i==0)
		return 0;
	else
		return 1;
}


REAL atan_approximate::b(long i)
{
	if (i==0)
		return x;
	else
		return i*i*x2/(4*(i*i)-1);
}


REAL atanh_approximate::a(long i)
{
	if (i==0)
		return 0;
	else
		return 1;
}


REAL atanh_approximate::b(long i)
{
	if (i==0)
		return x;
	else
		return 0-i*i*x2/(4*(i*i)-1);
}












